/***************************************************************************
                          SIMPLY POWERFUL TOOLKIT (SPTK)
                          CQueryGuard.h  -  description
                             -------------------
    begin                : Fir Jun 9 2006
    copyright            : (C) 2006-2008 by Alexey Parshin. All rights reserved.
    email                : alexeyp@gmail.com
 ***************************************************************************/

/***************************************************************************
  Redistribution and use in source and binary forms, with or without 
  modification, are permitted provided that the following conditions are met:

 * Redistributions of source code must retain the above copyright notice, 
    this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright notice, 
    this list of conditions and the following disclaimer in the documentation 
    and/or other materials provided with the distribution.
 * Neither the name of the <ORGANIZATION> nor the names of its contributors 
    may be used to endorse or promote products derived from this software 
    without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER 
  OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ***************************************************************************/

#ifndef __CQUERYGARD_H__
#define __CQUERYGARD_H__

#include <sptk3/cdatabase>

namespace sptk {

    /// @addtogroup Database Database Support
    /// @{

    /// @brief Service class to use with CQuery
    ///
    /// Manages the external CQuery object to make sure that it's automatically closed when it leaves the scope.

    class CQueryGuard {
        sptk::CQuery& m_query; ///< Wrapped query object
    public:
        /// @brief Constructor
        /// @param query sptk::CQuery, query to manage

        CQueryGuard(sptk::CQuery& query) : m_query(query) {
        }

        /// @brief Destructor

        ~CQueryGuard() {
            m_query.close();
        }

        /// @brief Opens managed query object

        void open() {
            m_query.open();
        }

        /// @brief Executes managed query object

        void exec() {
            m_query.exec();
        }

        /// @brief Fetches next data row from the managed query object

        void fetch() {
            m_query.fetch();
        }

        /// @brief Returns true if there is no more data rows to read

        bool eof() const {
            return m_query.eof();
        }

        /// @brief Allows to set query parameter by name
        /// @param paramName const char*, parameter name

        sptk::CParam& param(const char* paramName) {
            return m_query.param(paramName);
        }

        /// @brief Allows to set query parameter by parameter index
        /// @param paramIndex uint32_t, parameter index

        sptk::CParam& param(uint32_t paramIndex) {
            return m_query.param(paramIndex);
        }

        /// @brief Returns query field by the field name
        /// @param fieldName const char*, field name

        sptk::CField & operator [] (const char* fieldName) {
            return m_query[fieldName];
        }

        /// @brief Returns query field by field index
        /// @param fieldIndex uint32_t, field index

        sptk::CField & operator [] (uint32_t fieldIndex) {
            return m_query[fieldIndex];
        }

        /// @brief Returns managed query object

        operator sptk::CQuery& () {
            return m_query;
        }
    };

    /// @}
}

#endif
