/***************************************************************************
                          SIMPLY POWERFUL TOOLKIT (SPTK)
                          CThemeImageCollection.h  -  description
                             -------------------
    begin                : Sat Jun 28 2008
    copyright            : (C) 2003-2008 by Alexey Parshin. All rights reserved.
    email                : alexeyp@gmail.com
 ***************************************************************************/

/***************************************************************************
  Redistribution and use in source and binary forms, with or without 
  modification, are permitted provided that the following conditions are met:

 * Redistributions of source code must retain the above copyright notice, 
   this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright notice, 
   this list of conditions and the following disclaimer in the documentation 
   and/or other materials provided with the distribution.
 * Neither the name of the <ORGANIZATION> nor the names of its contributors 
   may be used to endorse or promote products derived from this software 
   without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER 
  OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ***************************************************************************/

#ifndef __CTHEMEIMAGECOLLECTION_H__
#define	__CTHEMEIMAGECOLLECTION_H__

#include <sptk3/cxml>
#include <sptk3/gui/CPngImage.h>

namespace sptk {

/// @addtogroup gui GUI Classes
/// @{

#define THMF_NORMAL 0
#define THMF_ACTIVE 1
#define THMF_HIGHLIGHTED 2

/// @brief Theme button type - defines which routine to use for which FLTK button type
enum CThemeButtonState { 
    THM_IMAGE_NORMAL = THMF_NORMAL, 
    THM_IMAGE_ACTIVE = THMF_ACTIVE, 
    THM_IMAGE_NORMAL_HIGHLITED = THMF_NORMAL | THMF_HIGHLIGHTED, 
    THM_IMAGE_ACTIVE_HIGHLITED = THMF_ACTIVE | THMF_HIGHLIGHTED, 
    THM_DEFAULT_FRAME };

#define MAX_BUTTON_STATES 5

/// @brief A collection of theme images
///
/// The images are stored for the different states of the object
class CThemeImageCollection {
    CPngImage*  m_images[MAX_BUTTON_STATES];
    CPngImage*  m_overlayImages[MAX_BUTTON_STATES];
    int         m_border[4];
public:
    CThemeImageCollection();
    ~CThemeImageCollection() { clear(); }
    
    void clear();
    
    void loadFromSptkTheme(const CStrings& objectNames);
    void loadFromGtkTheme(CXmlDoc& gtkTheme,std::string imagesXPath,std::string attribute="",std::string attributeValue="");

    int border(int ndx) const { return m_border[ndx]; }
    CPngImage* image(CThemeButtonState state) const; 
    CPngImage* overlayImage(CThemeButtonState state) const { return m_overlayImages[state]; }
};

/// @}
}

#endif

