/***************************************************************************
                          SIMPLY POWERFUL TOOLKIT (SPTK)
                          CGtkThemeLoader.h  -  description
                             -------------------
    begin                : Thu May 22 2008
    copyright            : (C) 2003-2008 by Alexey Parshin. All rights reserved.
    email                : alexeyp@gmail.com
 ***************************************************************************/

/***************************************************************************
  Redistribution and use in source and binary forms, with or without 
  modification, are permitted provided that the following conditions are met:

 * Redistributions of source code must retain the above copyright notice, 
    this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright notice, 
    this list of conditions and the following disclaimer in the documentation 
    and/or other materials provided with the distribution.
 * Neither the name of the <ORGANIZATION> nor the names of its contributors 
    may be used to endorse or promote products derived from this software 
    without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER 
  OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ***************************************************************************/

#ifndef __CGTKTHEMELOADER_H__
#define __CGTKTHEMELOADER_H__

#include <sptk3/cutils>
#include <sptk3/cxml>

namespace sptk {

    /// @addtogroup gui GUI Classes
    /// @{

    /// @brief This class parses GTK theme configuration file into XML
    class CGtkThemeParser {
        sptk::CXmlDoc   m_xml;          ///< Internal XML presentation of GTK theme config
        std::string     m_themeFolder;  ///< Path to the theme folder that contains gtkrc file

        /// @brief Reads parameter/value pair from the line of text
        /// @param row const std::string&, text row
        /// @param parentNode sptk::CXmlNoode* parentNode, a node to attach resulting node
        /// @returns resulting XML node
        sptk::CXmlNode* parseParameter(const std::string& row, sptk::CXmlNode* parentNode,bool createAttributes=false);
        
        /// @brief Parses a group of rows defining GTK image
        /// @param gtkrc const sptk::CStrings&, text of the GTK theme configuration file
        /// @param currentRow unsigned&, current row position in the GTK theme configuration file
        /// @param parentNode sptk::CXmlNoode* parentNode, a node to attach resulting node
        void parseImage(const sptk::CStrings& gtkrc, unsigned& currentRow, sptk::CXmlNode* parentNode);
        
        /// @brief Parses a group of rows defining GTK engine
        /// @param gtkrc const sptk::CStrings&, text of the GTK theme configuration file
        /// @param currentRow unsigned&, current row position in the GTK theme configuration file
        /// @param parentNode sptk::CXmlNoode* parentNode, a node to attach resulting node
        void parseEngine(const sptk::CStrings& gtkrc, unsigned& currentRow, sptk::CXmlNode* parentNode);
        
        /// @brief Parses a group of rows defining GTK style
        /// @param gtkrc const sptk::CStrings&, text of the GTK theme configuration file
        /// @param currentRow unsigned&, current row position in the GTK theme configuration file
        /// @param parentNode sptk::CXmlNoode* parentNode, a node to attach resulting node
        void parseStyle(const sptk::CStrings& gtkrc, unsigned& currentRow, sptk::CXmlNode* parentNode);
        
        /// @brief Parses a group of rows defining GTK theme
        /// @param gtkrc const sptk::CStrings&, text of the GTK theme configuration file
        void parse(const sptk::CStrings& gtkrc);
    public:

        /// @brief Default constructor
        CGtkThemeParser() {}
        
        /// @brief Loads GTK theme configuration
        void load(std::string themeName) throw (std::exception);
        
        sptk::CXmlDoc& xml() { return m_xml; }
        
        std::string themeFolder() { return m_themeFolder; }
    };

    /// @}
}

#endif
